#include "matrix.h"

#define TOL 1e-10

void jacobi(double** A, double* b, double* x, int n) {
   double err;
   double* r = (double*) malloc(n * sizeof(double));
   int iter=0;
   int i, j;
   double s;
   // Select initial solution
   for (i = 0; i < n; i++) {
     x[i] = 0.0;  
     r[i] = b[i];
   } 
   do {
     // Update solution
     for (i = 0; i < n; i++) {
       x[i] = r[i] / A[i][i] + x[i];
     }
     // Compute residual vector
     err = 0.0;
     for (i = 0; i < n; i++) {
       s = 0;
       for (j = 0; j < n; j++) {
         s += A[i][j] * x[j];
       }
       r[i] = b[i] - s;
       err += r[i] * r[i];
     }
     err = sqrt(err);
     iter++;
     // printf("iter=%d err=%e\n", iter, err);
   } while (err > TOL);
   printf("iter=%d err=%e\n", iter, err);
   free(r);
}

int main(int argc, char** argv) {
  --argc; ++argv;
  if (argc != 3 && argc != 4) {
    printf("Usage:\n"
           "  jacobi.bin A.mx b.mx x.mx [n_threads]\n"
           "Description:\n"
           "  - solves A x = b\n"
	   "  - A.mx and b.mx are the input files, x.mx is the output file\n"
           "  - if unspecified, n_threads defaults to 1\n");
    return 1;
  }

  int n_threads = argc == 4 ? atoi(argv[3]) : 1;
  if (n_threads <= 0) {
    printf("Invalid number of threads: %s\n", argv[3]);
    return 1;
  }

  const char* A_file = argv[0];
  const char* b_file = argv[1];
  const char* x_file = argv[2];
  int n, n_c1, n_c2;
  double** A = matrix_read(A_file, &n, &n_c1);

  if (!A) {
    printf("Error reading %s!\n", A_file);
    return 1;
  }

  if (n_c1 != n) {
    printf("%s does not hold a square matrix: %d x %d\n", A_file, n, n_c1);
    return 1;
  }
  double** b = matrix_read(b_file, &n_c1, &n_c2);
  if (n_c1 != 1) {
    printf("%s does not hold a vector: %d x %d\n", b_file, n_c1, n_c2);
    return 1;
  }
  if (n_c2 != n) {
    printf("%s holds an incompatible size vector: %d - expected  %d!\n", b_file, n_c2, n);
    return 1;
  }

  omp_set_num_threads(n_threads);

  printf("Solving A x = b using %d threads ...\n", n_threads);
  double** x = matrix_alloc(1, n);
  double t = - wall_time();
  jacobi(A, *b, *x, n);
  t += wall_time();
  printf("Time (s): %.3f\n", t);

  printf("Writing solution vector to %s ...\n", x_file);
  matrix_write(x_file, x, 1, n);
  matrix_free(A);
  matrix_free(b);

  return 0;
}
