#include "matrix.h"

double** matrix_alloc(int m, int n) {
  double* data = (double*) calloc(m * n, sizeof(double));

  if (data == NULL) {
    perror("calloc");
    return NULL;
  }

  double** matrix = (double**) calloc(m, sizeof(double*));

  if (matrix == NULL) {
    free(data);
    perror("calloc");
    return NULL;
  }

  for (int i = 0; i < m; i++) {
    matrix[i] = data;
    data += n;
  }

  return matrix;
}

void matrix_free(double** matrix) {
  free(matrix[0]);
  free(matrix);
}


void matrix_write(const char* file, double** matrix, int m, int n)  {
  FILE* fp = fopen(file, "w");
  if (!fp) { 
    perror(file);
    return;
  }
  fwrite(&m, 1, sizeof(int), fp);
  fwrite(&n, 1, sizeof(int), fp);
  fwrite(matrix[0], m * n, sizeof(double) , fp);
  fclose(fp);
}

double** matrix_read(const char* file, int* m, int* n)  {
  FILE* fp = fopen(file, "r");
  if (!fp) { 
    perror(file);
    return NULL;
  }
  fread(m, 1, sizeof(int), fp);
  fread(n, 1, sizeof(int), fp);

  double** matrix = matrix_alloc(*m, *n);
  fread(matrix[0], (*m) * (*n), sizeof(double) , fp);
  fclose(fp);

  return matrix;
}

double** matrix_gen(int m, int n, double a,  int seed) {
  double** matrix = matrix_alloc(m, n);

  if (!matrix) {
    return NULL;
  }

  srand(seed + m + n);
  for (int i=0; i < m; i++) {
    for (int j=0; j < n; j++) {
      double u = ((double) rand()) / RAND_MAX;
      double v = - a + 2 * u * a;
      if (a >= 3.0) {
        v = round(v); 
      }
      matrix[i][j] = v;
    }
  }

  return matrix;
}

void matrix_print(FILE* fp, double** matrix, int m, int n) {
  fprintf(fp, "== %d x %d matrix ==\n", m, n);
  for (int i = 0; i < m; i++) {
    fprintf(fp, "%3d:", i);
    for (int j = 0; j < n; j++) {
      fprintf(fp, " %12f", matrix[i][j]);
    }
    fputc('\n', fp);
  }
}

int matrix_compare(FILE* fp, double** a, double** b, int m, int n, double tol) {
  fprintf(fp, "== matrix comparison (%d x %d) ==\n", m, n);
  int total_errors = 0;
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      double delta = fabs(a[i][j] - b[i][j]);
      if (delta > tol) {
        fprintf(fp, "(%d,%d) : %f != %f [diff: %e]\n",  
                    i, j, a[i][j], b[i][j], delta);
        ++ total_errors;
      }
    }
  }
  fprintf(fp, "%d errors found.\n", total_errors);
  return total_errors;
}

void matrix_vmul(double** matrix, double* v, double *r, int m, int n) {
//  #pragma omp parallel 
  for (int i = 0; i < m; i++) {
    double s = 0;
    double* v_a = matrix[i];
    for (int j = 0; j < n; j++) {
      s += v_a[j] * v[j];
    }
    r[i] = s;
  }
}

#ifndef _OPENMP
double wall_time(void) {
  struct timeval tv;
  gettimeofday(&tv, 0);
  return (double) tv.tv_sec + ((double) tv.tv_usec * 1e-06);
}
#endif
